from typing import Any, Callable, Coroutine, Dict, List, Optional, Sequence, Type, Union

from . import routing
from .concurrency import AsyncExitStack
from .datastructures import Default, DefaultPlaceholder
from .encoders import DictIntStrAny, SetIntStr
from .exceptions import RequestValidationError
from .logger import logger
from .openapi.docs import (
    get_redoc_html,
    get_swagger_ui_html,
    get_swagger_ui_oauth2_redirect_html,
)
from .openapi.utils import get_openapi
from .params import Depends
from .types import DecoratedCallable
from .exceptions import HTTPException

from .exception_handlers import (
    http_exception_handler,
    request_validation_exception_handler,
)


# from starlette.applications import Starlette
# from starlette.datastructures import State
# from starlette.exceptions import HTTPException
# from starlette.middleware import Middleware
# from starlette.requests import Request
# from starlette.responses import HTMLResponse, JSONResponse, Response
# from starlette.routing import BaseRoute
# from starlette.types import ASGIApp, Receive, Scope, Send

from ..route import BaseRoute
from ..base import Request, Response, JSONResponse, HTMLResponse


class FastAPI:
    def __init__(
        self,
        *,
        debug: bool = False,
        routes: Optional[List[BaseRoute]] = None,
        title: str = "FastAPI",
        description: str = "",
        version: str = "0.1.0",
        openapi_url: Optional[str] = "/openapi.json",
        openapi_tags: Optional[List[Dict[str, Any]]] = None,
        servers: Optional[List[Dict[str, Union[str, Any]]]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        default_response_class: Type[Response] = Default(JSONResponse),
        docs_url: Optional[str] = "/docs",
        redoc_url: Optional[str] = "/redoc",
        swagger_ui_oauth2_redirect_url: Optional[str] = "/docs/oauth2-redirect",
        swagger_ui_init_oauth: Optional[Dict[str, Any]] = None,
        # middleware: Optional[Sequence[Middleware]] = None,
        exception_handlers: Optional[
            Dict[
                Union[int, Type[Exception]],
                Callable[[Request, Any], Coroutine[Any, Any, Response]],
            ]
        ] = None,
        # on_startup: Optional[Sequence[Callable[[], Any]]] = None,
        # on_shutdown: Optional[Sequence[Callable[[], Any]]] = None,
        terms_of_service: Optional[str] = None,
        contact: Optional[Dict[str, Union[str, Any]]] = None,
        license_info: Optional[Dict[str, Union[str, Any]]] = None,
        # openapi_prefix: str = "",
        root_path: str = "",
        root_path_in_servers: bool = True,
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        callbacks: Optional[List[BaseRoute]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        **extra: Any,
    ) -> None:
        self._debug: bool = debug
        # self.state: State = State()
        self.router: routing.APIRouter = routing.APIRouter(
            routes=routes,
            dependency_overrides_provider=self,
            prefix=root_path or "",
            # on_startup=on_startup,
            # on_shutdown=on_shutdown,
            default_response_class=default_response_class,
            dependencies=dependencies,
            callbacks=callbacks,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
            responses=responses,
        )
        self.exception_handlers: Dict[
            Union[int, Type[Exception]],
            Callable[[Request, Any], Coroutine[Any, Any, Response]],
        ] = (
            {} if exception_handlers is None else dict(exception_handlers)
        )
        self.exception_handlers.setdefault(HTTPException, http_exception_handler)
        self.exception_handlers.setdefault(
            RequestValidationError, request_validation_exception_handler
        )

        # self.user_middleware: List[Middleware] = (
        #     [] if middleware is None else list(middleware)
        # )
        # self.middleware_stack: ASGIApp = self.build_middleware_stack()

        self.title = title
        self.description = description
        self.version = version
        self.terms_of_service = terms_of_service
        self.contact = contact
        self.license_info = license_info
        self.servers = servers or []
        self.openapi_url = openapi_url
        self.openapi_tags = openapi_tags
        self.root_path = root_path
        self.root_path_in_servers = root_path_in_servers
        self.docs_url = docs_url
        self.redoc_url = redoc_url
        self.swagger_ui_oauth2_redirect_url = swagger_ui_oauth2_redirect_url
        self.swagger_ui_init_oauth = swagger_ui_init_oauth
        self.extra = extra
        self.dependency_overrides: Dict[Callable[..., Any], Callable[..., Any]] = {}

        self.openapi_version = "3.0.2"

        if self.openapi_url:
            assert self.title, "A title must be provided for OpenAPI, e.g.: 'My API'"
            assert self.version, "A version must be provided for OpenAPI, e.g.: '2.1.0'"
        self.openapi_schema: Optional[Dict[str, Any]] = None
        self.setup()

    @property
    def routes(self) -> List[BaseRoute]:
        return self.router.routes

    def openapi(self) -> Dict[str, Any]:
        if not self.openapi_schema:
            self.openapi_schema = get_openapi(
                title=self.title,
                version=self.version,
                openapi_version=self.openapi_version,
                description=self.description,
                terms_of_service=self.terms_of_service,
                contact=self.contact,
                license_info=self.license_info,
                routes=self.routes,
                tags=self.openapi_tags,
                servers=self.servers,
            )
        return self.openapi_schema

    def setup(self) -> None:
        if self.openapi_url:
            urls = (server_data.get("url") for server_data in self.servers)
            server_urls = {url for url in urls if url}

            def openapi(req: Request) -> JSONResponse:
                # root_path = req.scope.get("root_path", "").rstrip("/")
                # if root_path not in server_urls:
                #     if root_path and self.root_path_in_servers:
                #         self.servers.insert(0, {"url": root_path})
                #         server_urls.add(root_path)
                return JSONResponse(self.openapi())

            self.add_route(self.openapi_url, openapi, include_in_schema=False)

        if self.openapi_url and self.docs_url:

            def swagger_ui_html(req: Request) -> HTMLResponse:
                # root_path = req.scope.get("root_path", "").rstrip("/")
                root_path = self.root_path.rstrip("/")

                openapi_url = root_path + self.openapi_url
                oauth2_redirect_url = self.swagger_ui_oauth2_redirect_url
                if oauth2_redirect_url:
                    oauth2_redirect_url = root_path + oauth2_redirect_url
                return get_swagger_ui_html(
                    openapi_url=openapi_url,
                    title=self.title + " - Swagger UI",
                    oauth2_redirect_url=oauth2_redirect_url,
                    init_oauth=self.swagger_ui_init_oauth,
                )

            self.add_route(self.docs_url, swagger_ui_html, include_in_schema=False)

            if self.swagger_ui_oauth2_redirect_url:

                def swagger_ui_redirect(req: Request) -> HTMLResponse:
                    return get_swagger_ui_oauth2_redirect_html()

                self.add_route(
                    self.swagger_ui_oauth2_redirect_url,
                    swagger_ui_redirect,
                    include_in_schema=False,
                )
        if self.openapi_url and self.redoc_url:

            def redoc_html(req: Request) -> HTMLResponse:
                # root_path = req.scope.get("root_path", "").rstrip("/")
                root_path = self.root_path.rstrip("/")
                openapi_url = root_path + self.openapi_url
                return get_redoc_html(
                    openapi_url=openapi_url, title=self.title + " - ReDoc"
                )

            self.add_route(self.redoc_url, redoc_html, include_in_schema=False)

    # async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
    #     if self.root_path:
    #         scope["root_path"] = self.root_path
    #     if AsyncExitStack:
    #         async with AsyncExitStack() as stack:
    #             scope["fastapi_astack"] = stack
    #             await super().__call__(scope, receive, send)
    #     else:
    #         await super().__call__(scope, receive, send)  # pragma: no cover

    def add_route(
        self,
        path: str,
        endpoint: Callable,
        methods: List[str] = None,
        name: str = None,
        include_in_schema: bool = True,
    ) -> None:
        self.router.add_route(
            path=self.root_path + "/" + path.strip("/"),
            endpoint=endpoint,
            methods=methods,
            name=name,
            include_in_schema=include_in_schema,
        )

    def add_api_route(
        self,
        path: str,
        endpoint: Callable[..., Coroutine[Any, Any, Response]],
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        methods: Optional[List[str]] = None,
        operation_id: Optional[str] = None,
        response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(
            JSONResponse
        ),
        name: Optional[str] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.router.add_api_route(
            path,
            endpoint=endpoint,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            dependencies=dependencies,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            methods=methods,
            operation_id=operation_id,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            openapi_extra=openapi_extra,
        )

    def api_route(
        self,
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        methods: Optional[List[str]] = None,
        operation_id: Optional[str] = None,
        response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        def decorator(func: DecoratedCallable) -> DecoratedCallable:
            self.router.add_api_route(
                path,
                func,
                response_model=response_model,
                status_code=status_code,
                tags=tags,
                dependencies=dependencies,
                summary=summary,
                description=description,
                response_description=response_description,
                responses=responses,
                deprecated=deprecated,
                methods=methods,
                operation_id=operation_id,
                response_model_include=response_model_include,
                response_model_exclude=response_model_exclude,
                response_model_by_alias=response_model_by_alias,
                response_model_exclude_unset=response_model_exclude_unset,
                response_model_exclude_defaults=response_model_exclude_defaults,
                response_model_exclude_none=response_model_exclude_none,
                include_in_schema=include_in_schema,
                response_class=response_class,
                name=name,
                openapi_extra=openapi_extra,
            )
            return func

        return decorator

    def add_api_websocket_route(
        self, path: str, endpoint: Callable[..., Any], name: Optional[str] = None
    ) -> None:
        self.router.add_api_websocket_route(path, endpoint, name=name)

    def websocket(
        self, path: str, name: Optional[str] = None
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        def decorator(func: DecoratedCallable) -> DecoratedCallable:
            self.add_api_websocket_route(path, func, name=name)
            return func

        return decorator

    def include_router(
        self,
        router: routing.APIRouter,
        *,
        prefix: str = "",
        tags: Optional[List[str]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        default_response_class: Type[Response] = Default(JSONResponse),
        callbacks: Optional[List[BaseRoute]] = None,
    ) -> None:
        self.router.include_router(
            router,
            prefix=prefix,
            tags=tags,
            dependencies=dependencies,
            responses=responses,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
            default_response_class=default_response_class,
            callbacks=callbacks,
        )

    def get(
        self,
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        callbacks: Optional[List[BaseRoute]] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.router.get(
            path,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            dependencies=dependencies,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            operation_id=operation_id,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            callbacks=callbacks,
            openapi_extra=openapi_extra,
        )

    def put(
        self,
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        callbacks: Optional[List[BaseRoute]] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.router.put(
            path,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            dependencies=dependencies,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            operation_id=operation_id,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            callbacks=callbacks,
            openapi_extra=openapi_extra,
        )

    def post(
        self,
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        callbacks: Optional[List[BaseRoute]] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.router.post(
            path,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            dependencies=dependencies,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            operation_id=operation_id,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            callbacks=callbacks,
            openapi_extra=openapi_extra,
        )

    def delete(
        self,
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        callbacks: Optional[List[BaseRoute]] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.router.delete(
            path,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            dependencies=dependencies,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            operation_id=operation_id,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            callbacks=callbacks,
            openapi_extra=openapi_extra,
        )

    def options(
        self,
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        callbacks: Optional[List[BaseRoute]] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.router.options(
            path,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            dependencies=dependencies,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            operation_id=operation_id,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            callbacks=callbacks,
            openapi_extra=openapi_extra,
        )

    def head(
        self,
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        callbacks: Optional[List[BaseRoute]] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.router.head(
            path,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            dependencies=dependencies,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            operation_id=operation_id,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            callbacks=callbacks,
            openapi_extra=openapi_extra,
        )

    def patch(
        self,
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        callbacks: Optional[List[BaseRoute]] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.router.patch(
            path,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            dependencies=dependencies,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            operation_id=operation_id,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            callbacks=callbacks,
            openapi_extra=openapi_extra,
        )

    def trace(
        self,
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[str]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        callbacks: Optional[List[BaseRoute]] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.router.trace(
            path,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            dependencies=dependencies,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            operation_id=operation_id,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            callbacks=callbacks,
            openapi_extra=openapi_extra,
        )



from datetime import datetime
from hashlib import md5
from typing import List, Optional

from django.http.response import Http404

try:
    from django import setup as django_setup
except ImportError:

    def django_setup():
        pass


from django.http import HttpResponseRedirect
from django.conf.urls import url
from pydantic import BaseModel, Field
from django_mini_fastapi import (
    OpenAPI,
    Path,
    Body,
    Query,
    Form,
    File,
    UploadFile,
    Cookie,
    Header,
    Request,
    Session,
    Response,
    Depends,
)


"""
A minimal working demo of Django App
"""

DEBUG = True
SECRET_KEY = "canukeepasecret"
ROOT_URLCONF = __name__

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
    }
}

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
]

SESSION_ENGINE = "django.contrib.sessions.backends.signed_cookies"


django_setup()


def redirect_to_doc(request):
    return HttpResponseRedirect(
        "/intro/rapidoc#tag--1.-Setup-your-first-OpenAPI-endpoint"
    )


api = OpenAPI(
    title="OpenAPI Test",
    version="0.1",
    description="Just a Test",
    root_path="/intro",
)

urlpatterns = [
    url(r"^$", redirect_to_doc),
    api.as_django_url_pattern(),
]


class IntroResponse1(BaseModel):
    arg1: str
    arg2: str


class IntroResponse2(BaseModel):
    arg1: str = Field(min_length=3, max_length=10)
    arg2: int = Field(ge=0, le=10)
    arg3: bool = Field(False)


@api.get(
    "/basic_get_request",
    tags=["1. Setup your first OpenAPI endpoint"],
    summary="Get start & create a simple http GET route",
)
def basic_get_request():
    """
    For start using django-mini-fastapi, in your Django project

    * import OpenAPI from django_mini_fastapi
    * create an OpenAPI object
    * put it into your urlpatterns
    * start define api endpoints using api OpenAPI object


    ```python
    from django_mini_fastapi import OpenAPI

    api = OpenAPI(
        title='OpenAPI Test',
        version='0.1',
        description='Just a Test',
        root_path='/intro'
    )

    urlpatterns = [
        api.as_django_url_pattern()
    ]

    @api.get('/basic_get_request')
    def basic_get_request():
        return {'hello': 'world'}
    ```
    """
    return {"hello": "world"}


@api.get(
    "/test_path_and_query_parameters/{arg1}",
    tags=["1. Basic HTTP requests"],
    summary="Define path & query parameters",
    response_model=IntroResponse1,
)
def test_path_and_query_parameters(
    arg1,
    arg2,
):
    """
    Use same arg name as the one in path for receiving path args
    For those args which names not matched path arg names, will be parsed as query parameter

    ```python
    from django_mini_fastapi import Path

    @api.get('/test_path_and_query_parameters/{arg1}')
    def test_path_and_query_parameters(arg1, arg2):
        return dict(arg1=arg1, arg2=arg2)
    ```
    """
    return dict(arg1=arg1, arg2=arg2)


@api.get(
    "/basic_check_on_path_or_query_parameter/{arg1}",
    tags=["1. Basic HTTP requests"],
    summary="Validate query string parameters",
    response_model=IntroResponse1,
)
def basic_check_on_path_or_query_parameter(
    arg1: int = Path(..., ge=0), arg2: str = Query("default", min_length=3)
):
    """
    Use `Query()` or `Path()` to tell API how to parse and check constraints of parameters

    ```python
    from django_mini_fastapi import Query, Path

    @api.get('/basic_check_on_path_or_query_parameter/{arg1}')
    def basic_check_on_path_or_query_parameter(arg1: int=Path(..., ge=0), arg2: str=Query('default')):
        return dict(arg1=arg1, arg2=arg2)
    ```
    """
    return dict(arg1=arg1, arg2=arg2)


@api.get(
    "/get_request_with_json_schema_query_args",
    tags=["1. Basic HTTP requests"],
    summary="Auto parameter validation via JSON schema fields",
    response_model=IntroResponse2,
)
def get_request_with_json_schema_query_args(
    arg1: str = Query(..., min_length=3, max_length=10),
    arg2: int = Query(..., ge=0, le=10),
    arg3: bool = Query(False),
):
    """
    ```python
    from django_mini_fastapi import Query

    @api.get('/get_request_with_json_schema_query_args')
    def get_request_with_json_schema_query_args(
        arg1: str = Query(..., min_length=3, max_length=10),
        arg2: int = Query(..., ge=0, le=10),
        arg3: bool = Query(False),
    ):
        return dict(arg1=arg1, arg2=arg2, arg3=arg3)
    ```
    """
    return dict(arg1=arg1, arg2=arg2, arg3=arg3)


@api.post(
    "/post_request_with_json_schema_form_args",
    tags=["1. Basic HTTP requests"],
    summary="Define Form parameters",
    response_model=IntroResponse2,
)
def post_request_with_json_schema_form_args(
    arg1: str = Form(..., min_length=3, max_length=10),
    arg2: int = Form(..., ge=0, le=10),
    arg3: bool = Form(False),
):
    """
    Now we use the same JSON schema field definitions, but in Form() format.

    ```python
    from django_mini_fastapi import Form
    from django_mini_fastapi.schema import StringField, NumberField, BooleanField

    @api.post('/post_request_with_json_schema_form_args')
    def post_request_with_json_schema_form_args(
        arg1: str = Form(..., min_length=3, max_length=10),
        arg2: int = Form(..., ge=0, le=10),
        arg3: bool = Form(False),
    ):
        return dict(arg1=arg1, arg2=arg2, arg3=arg3)
    ```
    """
    return dict(arg1=arg1, arg2=arg2, arg3=arg3)


@api.post(
    "/post_request_file_upload",
    tags=["1. Basic HTTP requests"],
    summary="Define File Upload",
)
def post_request_file_upload(
    upload_file: UploadFile = File(...),
    md5_hash: Optional[str] = Form(None, description="md5 of uploaded file"),
):
    """
    Now let's try building an endpoint for user file upload.

    ```python
    from hashlib import md5
    from django_mini_fastapi import UploadedFile, Form, File

    @api.post('/post_request_file_upload')
    def post_request_file_upload(
        upload_file: UploadedFile = File(...),
        md5_hash: Optional[str] = Form(None, description="md5 of uploaded file"),
    ):
        return {
            'submitted_md5': md5_hash,
            'file': {
                'name': upload_file.name,
                'size': upload_file.size,
                'md5': md5(upload_file.read()).hexdigest,
            },
        }
    ```
    """
    return {
        "submitted_md5": md5_hash,
        "file": {
            "name": upload_file.name,
            "size": upload_file.size,
            "md5": md5(upload_file.read()).hexdigest(),
        },
    }


class SamplePayload(BaseModel):
    arg1: str = Field(..., min_length=3, max_length=10)
    arg2: int = Field(..., ge=0, le=10)
    arg3: bool = Field(..., default_value=False)


class SampleResponse(BaseModel):
    obj: SamplePayload
    ary: List[SamplePayload]


@api.post(
    "/post_request_with_json_schema_body",
    tags=["1. Basic HTTP requests"],
    summary="Define body parameters via JSON schema model",
    response_model=SampleResponse,
)
def post_request_with_json_schema_body(payload: SamplePayload):
    """
    The JSON schema fields could also be used for describing JSON body format,
    all you need is declearing a class inherited from BaseModel class.

    ```python
    from pydantic import BaseModel, Field

    class SamplePayload(BaseModel):
        arg1: str = Field(..., min_length=3, max_length=10)
        arg2: int = Field(..., ge=0, le=10)
        arg3: bool = Field(..., default_value=False)


    class SampleResponse(BaseModel):
        obj: SamplePayload
        ary: List[SamplePayload]


    @api.post(
        '/post_request_with_json_schema_body',
        response_model=SampleResponse,  # you can also put json schema model here for declearing response model
    )
    def post_request_with_json_schema_body(
        payload: SamplePayload,
    ):
        return payload
    ```
    """
    return {"obj": payload, "ary": [payload, payload]}


@api.post(
    "/accessing_special_raw_variables",
    tags=["1. Basic HTTP requests"],
    summary="Some special variables",
)
def accessing_special_raw_variables(
    request: Request, session: Session, response: Response
):
    """
    You could access some special variables via your function parameter type.

    * `django_mini_fast_api.Request`

      * For accessing the raw django `HttpRequest` object

    * `django_mini_fast_api.Session`

      * django session object binded on request

    * ``django_mini_fast_api.Response``

      * A temp empty response for manipulating/overriding status code / headers / cookies on the final response object

    ```python
    @api.get(
        '/accessing_special_raw_variables',
    )
    def accessing_special_raw_variables(request: Request, session: Session, response: Response):

        current_dt_str = str(datetime.now())
        response.set_cookie("test_cookie", current_dt_str)
        response["X-TEST-HEADER"] = current_dt_str

        session["ts"] = current_dt_str
        session.save()

        return {
            'request': request,
            'session': session,
        }
    ```
    """
    current_dt_str = str(datetime.now())
    response.set_cookie("test_cookie", current_dt_str)
    response["X-TEST-HEADER"] = current_dt_str

    session["ts"] = current_dt_str
    session.save()

    return {
        "request": repr(request),
        "session": session,
    }


@api.post(
    "/other_argument_data_sources",
    tags=["1. Basic HTTP requests"],
    summary="Other argument data sources",
)
def other_argument_data_sources(
    test_cookie=Cookie(...), referer=Header(...), x_api_token=Header(...)
):
    """
    You can also get your request arguments from other data sources.

    Like: `Header()`, `Cookie()`

    * *Note: While accessing via `Header()`, all variable name will converted to uppercase as http header key name*
    * *e.g.: `content_type` -> `CONTENT_TYPE`

    ```python
    @api.post(
        '/other_argument_data_sources',
    )
    def other_argument_data_sources(
        test_cookie=Cookie(...), referer=Header(...), x_api_token=Header(...)
    ):
        return {
            "test_cookie": test_cookie,
            "x_api_token": x_api_token,
            "referer": referer,
        }
    ```
    """
    return {
        "test_cookie": test_cookie,
        "x_api_token": x_api_token,
        "referer": referer,
    }


class Pagination(BaseModel):
    page: int
    limit: int


def get_pagination(page: int = Query(1, ge=1), limit: int = Query(20, ge=0)):
    return Pagination(page=page, limit=limit)


@api.get(
    "/basic_dependencies",
    tags=["2. Dependencies"],
    summary="Basic dependencies sample",
)
def basic_dependencies(pagination: Pagination = Depends(get_pagination)):
    """
    For common operations across multiple API endpoints, you could use Depends() for merging them into single dependency functions

    For example:
    You need to parse pagination info on each API endpoint, you could write a `get_pagination` function and let your API endpoint `Depends()` on it.

    ```python
    class Pagination(BaseModel):
        page: int
        limit: int

    def get_pagination(page: int = Query(1, ge=1), limit: int = Query(20, ge=0)):
        return Pagination(page=page, limit=limit)

    @api.get("/basic_dependencies")
    def basic_dependencies(
        pagination: Pagination = Depends(get_pagination)
    ):
        return {"pagination": pagination}
    ```
    """
    return {"pagination": pagination}


def check_api_key(x_api_key: str = Header(...)):
    if x_api_key != "kamehame":
        raise Http404


@api.get(
    "/non_return_dependencies",
    tags=["2. Dependencies"],
    summary="Dependencies without return value",
    dependencies=[Depends(check_api_key)],
)
def non_return_dependencies():
    """
    For dependencies only need to be executed but needn't to return any values, you could declear it in API endpoint declaration function with `dependencies=` keyword argument

    For example:
    You need to check API key on every API endpoint, you could write a `check_api_key` function put into API endpoint's (even router) dependencies

    ```python
    def check_api_key(x_api_key: str = Header(...)):
        if x_api_key != "kamehame":
            raise Http404

    @api.get("/non_return_dependencies", dependencies=[Depends(check_api_key)],)
    def non_return_dependencies():
        return {"msg": "You have correct API key!"}
    ```
    """
    return {"msg": "You have correct API key!"}



from inspect import isclass, isfunction

from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Set,
    Union,
)
from string import Formatter
import re

from pydantic import BaseModel, Field

from django.utils.encoding import force_text

from .base import Request, Response

from .enums import AllowHttpMethodEnum


class RouteConfig(BaseModel):
    route_path: str = Field(..., min_length=1, regex=r"^/")
    allow_methods: List[AllowHttpMethodEnum] = Field(min_items=1)
    summary: Optional[str]
    description: Optional[str]
    tags: List[str]


class RoutePath(object):
    def __init__(self, route_path: str):
        route_path = force_text(route_path)
        assert route_path.startswith("/")

        self.org_route_path = route_path
        # route_path = route_path[1:]
        self.route_path = route_path[:-1] if route_path.endswith("/") else route_path

        self.key_set = set()
        re_segs = []
        fmt = Formatter()
        for prefix, key, fmt_spec, conversion in fmt.parse(self.route_path):
            if key is None:
                re_segs.append(prefix)
                continue

            assert key and not fmt_spec and not conversion, (
                "fail parsing parameter from path: " + self.route_path
            )
            assert key not in self.key_set, "duplicated key {}".format(key)
            self.key_set.add(key)
            re_segs.append("{prefix}(?P<{key}>[^/]+)".format(prefix=prefix, key=key))

        # print('regex: ' + ''.join(re_segs))
        self.regex = re.compile("^" + "".join(re_segs) + "$")

    def parse(self, request_path) -> Optional[Dict[str, str]]:
        if not request_path.startswith("/"):
            request_path = "/" + request_path

        match = self.regex.match(request_path)
        if not match:
            return None

        return match.groupdict()


def get_name(endpoint: Callable) -> str:
    if isfunction(endpoint) or isclass(endpoint):
        return endpoint.__name__

    return endpoint.__class__.__name__


class BaseRoute:
    def __init__(
        self,
        path: str,
        endpoint: Callable,
        *,
        name: Optional[str] = None,
        methods: Optional[Union[Set[str], List[str]]] = None,
        include_in_schema: bool = True,
    ) -> None:
        self.path = path
        self.endpoint: Callable = endpoint
        self.name = name if name is not None else get_name(endpoint)
        self.path_parser: RoutePath = RoutePath(path)

        if methods is None:
            methods = ["GET"]
        self.methods: Set[str] = set([method.upper() for method in methods])

        self.include_in_schema: bool = include_in_schema

        def route_handler(request):
            return self.endpoint(request)

        self.route_handler: Callable[[Request], Response] = route_handler

    def match_path(self, request_path) -> Dict[str, str]:
        return self.path_parser.parse(request_path)

    def check_method_allowed(self, method):
        return method in self.methods

    def __call__(self, request: Request) -> Response:
        return self.route_handler(request)


class Router:
    def __init__(
        self,
        routes: Sequence[BaseRoute] = None,
        redirect_slashes: bool = True,
    ):
        self.routes = [] if routes is None else list(routes)
        self.redirect_slashes = redirect_slashes

    def add_route(
        self,
        path: str,
        endpoint: Callable,
        methods: List[str] = None,
        name: str = None,
        include_in_schema: bool = True,
    ) -> None:
        route = BaseRoute(
            path=path,
            endpoint=endpoint,
            methods=methods,
            name=name,
            include_in_schema=include_in_schema,
        )
        self.routes.append(route)



from typing import Any, Callable, Iterable, Type
from django.http import (
    HttpRequest as Request,
    HttpResponse as Response,
    JsonResponse,
)
from django.core.files.uploadedfile import UploadedFile

try:
    from django.http.request import HttpHeaders as Headers
except ImportError:
    Headers = dict
from django.http.request import QueryDict as QueryParams
from django.http.request import QueryDict as FormData
from django.contrib.sessions.backends.base import SessionBase as Session


class JSONResponse(JsonResponse):
    media_type = "application/json"


class HTMLResponse(Response):
    media_type = "text/html"


class TempResponse(Response):
    def __init__(self) -> None:
        super().__init__()
        # set empty status_code
        self.status_code = None

        # reset headers
        for k in ("headers", "_headers"):
            headers = getattr(self, k, None)
            if headers:
                setattr(self, k, headers.__class__({}))


class UploadFile(UploadedFile):
    @classmethod
    def __get_validators__(cls: Type["UploadFile"]) -> Iterable[Callable[..., Any]]:
        yield cls.validate

    @classmethod
    def validate(cls: Type["UploadFile"], v: Any) -> Any:
        if not isinstance(v, UploadedFile):
            raise ValueError(f"Expected UploadedFile, received: {type(v)}")
        return v



from .encoders import jsonable_encoder
from .exceptions import RequestValidationError, HTTPException

# from starlette.exceptions import HTTPException
# from starlette.requests import Request
# from starlette.responses import JSONResponse
# from starlette.status import HTTP_422_UNPROCESSABLE_ENTITY

from http import HTTPStatus

HTTP_422_UNPROCESSABLE_ENTITY = int(HTTPStatus.UNPROCESSABLE_ENTITY)

from ..base import Request, JSONResponse


def http_exception_handler(request: Request, exc: HTTPException) -> JSONResponse:
    headers = getattr(exc, "headers", None)
    if headers:
        return JSONResponse(
            {"detail": exc.detail}, status=exc.status_code, headers=headers
        )
    else:
        return JSONResponse({"detail": exc.detail}, status=exc.status_code)


def request_validation_exception_handler(
    request: Request, exc: RequestValidationError
) -> JSONResponse:
    return JSONResponse(
        status=HTTP_422_UNPROCESSABLE_ENTITY,
        data={"detail": jsonable_encoder(exc.errors())},
    )



from enum import Enum


class StrEnum(str, Enum):
    pass


class AllowHttpMethodEnum(StrEnum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"
    HEAD = "HEAD"
    TRACE = "TRACE"
    OPTION = "OPTION"



from typing import Any, Callable, Coroutine, Dict, List, Optional, Sequence, Type, Union

from django.conf.urls import url
from django.views.decorators.csrf import csrf_exempt
from django.http import Http404, HttpResponseNotAllowed

from .fastapi import FastAPI
from .fastapi.params import Depends
from .fastapi.exceptions import HTTPException
from .fastapi.datastructures import Default
from .base import HTMLResponse, Request, Response, JSONResponse
from .route import BaseRoute

import logging

_logger = logging.getLogger(__name__)


RAPIDOC_PAGE_TPL = """
<!doctype html> <!-- Important: must specify -->
<html>
<head>
<title>{title} - RapiDoc</title>
<meta charset="utf-8"> <!-- Important: rapi-doc uses utf8 charecters -->
<script type="module" src="https://unpkg.com/rapidoc/dist/rapidoc-min.js"></script>
</head>
<body>
<rapi-doc
    spec-url="{openapi_url}"
    sort-endpoints-by="method"
    render-style="read"
> </rapi-doc>
</body>
</html>
"""


class OpenAPI(FastAPI):
    def __init__(
        self,
        *,
        debug: bool = False,
        routes: Optional[List[BaseRoute]] = None,
        title: str = "Django mini FastAPI",
        description: str = "",
        version: str = "0.1.0",
        openapi_url: Optional[str] = "/openapi.json",
        openapi_tags: Optional[List[Dict[str, Any]]] = None,
        servers: Optional[List[Dict[str, Union[str, Any]]]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        default_response_class: Type[Response] = Default(JSONResponse),
        docs_url: Optional[str] = "/docs",
        redoc_url: Optional[str] = "/redoc",
        rapidoc_url: Optional[str] = "/rapidoc",
        swagger_ui_oauth2_redirect_url: Optional[str] = "/docs/oauth2-redirect",
        swagger_ui_init_oauth: Optional[Dict[str, Any]] = None,
        exception_handlers: Optional[
            Dict[
                Union[int, Type[Exception]],
                Callable[[Request, Any], Coroutine[Any, Any, Response]],
            ]
        ] = None,
        terms_of_service: Optional[str] = None,
        contact: Optional[Dict[str, Union[str, Any]]] = None,
        license_info: Optional[Dict[str, Union[str, Any]]] = None,
        root_path: str = "",
        root_path_in_servers: bool = True,
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        callbacks: Optional[List[BaseRoute]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        **extra: Any
    ) -> None:
        super().__init__(
            debug=debug,
            routes=routes,
            title=title,
            description=description,
            version=version,
            openapi_url=openapi_url,
            openapi_tags=openapi_tags,
            servers=servers,
            dependencies=dependencies,
            default_response_class=default_response_class,
            docs_url=docs_url,
            redoc_url=redoc_url,
            swagger_ui_oauth2_redirect_url=swagger_ui_oauth2_redirect_url,
            swagger_ui_init_oauth=swagger_ui_init_oauth,
            exception_handlers=exception_handlers,
            terms_of_service=terms_of_service,
            contact=contact,
            license_info=license_info,
            root_path=root_path,
            root_path_in_servers=root_path_in_servers,
            responses=responses,
            callbacks=callbacks,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
            **extra
        )

        self.rapidoc_url = rapidoc_url

        if self.openapi_url and self.rapidoc_url:

            def rapi_doc_html(req: Request) -> HTMLResponse:
                root_path = self.root_path.rstrip("/")
                openapi_url = root_path + self.openapi_url

                return HTMLResponse(
                    RAPIDOC_PAGE_TPL.format(title=self.title, openapi_url=openapi_url)
                )

            self.add_route(self.rapidoc_url, rapi_doc_html, include_in_schema=False)

    def as_django_url_pattern(self):
        return url(
            "^{prefix_path}/(?P<route_path>.*)".format(
                prefix_path=self.root_path.strip("/")
            ),
            self.as_django_view(),
        )

    def as_django_view(self):
        @csrf_exempt
        def dispatcher(request: Request, route_path: str):
            route_path = self.root_path + "/" + route_path.strip("/")

            matched_route = None
            matched_route_path_kwargs = None
            method_not_allowed_routes: List[BaseRoute] = []

            try:
                for route in self.router.routes:
                    path_kwargs: Optional[Dict[str, str]] = route.match_path(route_path)

                    # path regex not matched
                    if path_kwargs is None:
                        continue

                    # found 1st full matched route, break here
                    if route.check_method_allowed(request.method):
                        matched_route = route
                        matched_route_path_kwargs = path_kwargs
                        break
                    else:
                        method_not_allowed_routes.append(route)
                else:
                    # no break after scanned all routes
                    if method_not_allowed_routes:
                        raise HTTPException(405)
                    else:
                        raise HTTPException(404)

                request.path_kwargs = matched_route_path_kwargs
                return matched_route(request)

            except Exception as e:
                exc_handler = self.exception_handlers.get(type(e))
                if exc_handler:
                    return exc_handler(request, e)
                raise e

        return dispatcher

    def add_exception_handler(self, exc_cls: Type[Exception], fn: Callable):
        self.exception_handlers[exc_cls] = fn

    def exception_handler(self, exc_cls: Type[Exception]) -> Callable:
        def _decorated(fn: Callable) -> Callable:
            self.add_exception_handler(exc_cls=exc_cls, fn=fn)

        return _decorated



import http
from typing import Any, Dict, Optional, Sequence, Type

from pydantic import BaseModel, ValidationError, create_model
from pydantic.error_wrappers import ErrorList

# from starlette.exceptions import HTTPException as StarletteHTTPException

RequestErrorModel: Type[BaseModel] = create_model("Request")
WebSocketErrorModel: Type[BaseModel] = create_model("WebSocket")


class FastAPIError(RuntimeError):
    """
    A generic, FastAPI-specific error.
    """


class RequestValidationError(ValidationError):
    def __init__(self, errors: Sequence[ErrorList], *, body: Any = None) -> None:
        self.body = body
        super().__init__(errors, RequestErrorModel)


class WebSocketRequestValidationError(ValidationError):
    def __init__(self, errors: Sequence[ErrorList]) -> None:
        super().__init__(errors, WebSocketErrorModel)


class HTTPException(Exception):
    def __init__(self, status_code: int, detail: str = None) -> None:
        if detail is None:
            detail = http.HTTPStatus(status_code).phrase
        self.status_code = status_code
        self.detail = detail

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}(status_code={self.status_code!r}, detail={self.detail!r})"



"""Minimal FastAPI implementation for Django"""

__version__ = "0.1.7"


from .api import OpenAPI as OpenAPI
from .base import (
    Request as Request,
    Response as Response,
    UploadFile as UploadFile,
    Session as Session,
)
from .fastapi.param_functions import (
    Path as Path,
    Query as Query,
    Form as Form,
    File as File,
    Body as Body,
    Header as Header,
    Cookie as Cookie,
    Depends as Depends,
)
from .fastapi.routing import APIRoute as APIRoute
from .fastapi.routing import APIRouter as APIRouter

from .fastapi.exceptions import HTTPException as HTTPException



from django.test import TestCase


class TestApiRouter(TestCase):
    def test_path_args(self):
        resp = self.client.get(
            "/intro/test_path_and_query_parameters", dict(arg1="abc", arg2="def")
        )
        self.assertEqual(resp.status_code, 404)

        resp = self.client.get(
            "/intro/test_path_and_query_parameters/abc", dict(arg2="def")
        )
        resp_json = resp.json()
        self.assertEqual(resp_json["arg1"], "abc")
        self.assertEqual(resp_json["arg2"], "def")

    def test_query_args_with_schema(self):
        resp = self.client.get(
            "/intro/basic_check_on_path_or_query_parameter/abc",
            dict(arg1="abc", arg2="def"),
        )
        self.assertEqual(resp.status_code, 422)

        resp_json = self.client.get(
            "/intro/get_request_with_json_schema_query_args",
            dict(arg1="abc", arg2="5", arg3="1"),
        ).json()
        self.assertEqual(resp_json["arg1"], "abc")
        self.assertEqual(resp_json["arg2"], 5)
        self.assertEqual(resp_json["arg3"], True)



# from fastapi.security.base import SecurityBase
class SecurityBase:
    pass


# from fastapi.security.oauth2 import OAuth2, SecurityScopes
class OAuth2:
    pass


# from fastapi.security.oauth2 import OAuth2, SecurityScopes
class SecurityScopes:
    def __init__(self, *, scopes) -> None:
        self.scopes = scopes


# from fastapi.security.open_id_connect_url import OpenIdConnect
class OpenIdConnect:
    pass


# from starlette.concurrency import iterate_in_threadpool as iterate_in_threadpool  # noqa
async def iterate_in_threadpool(*args, **kwargs):
    pass


# from starlette.concurrency import run_in_threadpool as run_in_threadpool  # noqa
async def run_in_threadpool(*args, **kwargs):
    pass


# from starlette.concurrency import (  # noqa
#     run_until_first_complete as run_until_first_complete,
# )
async def run_until_first_complete(*args, **kwargs):
    pass


# from starlette.websockets import WebSocket
class WebSocket:
    pass


# from starlette.requests import HTTPConnection
class HTTPConnection:
    pass


# from starlette.background import BackgroundTasks
class BackgroundTasks:
    pass



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['django_mini_fastapi',
 'django_mini_fastapi.fastapi',
 'django_mini_fastapi.fastapi.dependencies',
 'django_mini_fastapi.fastapi.openapi']

package_data = \
{'': ['*']}

setup(name='django-mini-fastapi',
      version='0.1.7',
      description='Minimal FastAPI implementation for Django',
      author=None,
      author_email='toki <toki.kanno@gmail.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      python_requires='>=3.6',
     )



import re
from django.http import JsonResponse


def json_response(data, status_code=200):
    resp = JsonResponse(data)
    resp.status_code = status_code
    return resp


def get_full_qualified_name(obj):
    return f"{obj.__module__}.{obj.__qualname__}"


def generate_operation_id_for_path(*, name: str, path: str, method: str) -> str:
    operation_id = name + path
    operation_id = re.sub("[^0-9a-zA-Z_]", "_", operation_id)
    operation_id = operation_id + "_" + method.lower()
    return operation_id
